/*======================================================================*/
/*                                                                      */
/*      dccpp_specific.c                                                */
/*      This module contains functions related to specific local        */
/*      handling aimed at patching major missing features in the DCC++  */
/*      station.                                                        */
/*      This includes loco data base handling, and accessory data base  */
/*      handling.                                                       */
/*======================================================================*/
/*      J.P.PILLOU      13/03/2018                                     */
/*======================================================================*/
/*                                                                      */
/*======================================================================*/
/*      J.P.PILLOU      2013/01/22                                      */
/*======================================================================*/
/*======================================================================*/

#include <windows.h>
#include <stdio.h>
#include "DDGI_Generic_def.h"
#include "dccpp_def.h"
#include "dccpp_extern.h"
#include "dccpp_specific.h"

/*======================================================================*/
/*                         FUNCTION DECLARATIONS                        */
/*======================================================================*/


/* LOCOS UTILITIES */

BOOL    DCPS_EmergencyStopAllLocos( DCC_CMD_TAG         *pCdmTag );

BOOL    DCPS_RestartAllLocos( void );


unsigned char
        DCPS_SetFxByte(         unsigned int            nFuncValue, 
                                unsigned int            nFuncMask );

int     DCPS_UpdateLocoParameters( int                  nAddress, 
                                int                     nSpeed, 
                                SPEED_STEP_MODE         nSpeedStepMode,
                                BOOL                    bForward, 
                                unsigned int            nF0F12,
                                unsigned int            nF0F12Mask );

void    DCPS_UpdateLocoSpeed(   int                     nAddress, 
                                int                     nSpeed, 
                                SPEED_STEP_MODE         nSpeedStepMode );

void    DCPS_UpdateLocoFx(      int                     nAddress, 
                                unsigned int            nF0F12,
                                unsigned int            nF0F12Mask );

void    DCPS_SetEmergencyStop(  int                     nAddress );

BOOL    DCPS_GetLocoInfo(       pDGI_GENERIC_RCV_DATA   pGenericRcvData, 
                                int                     nAddress );

BOOL    DCPS_GetLocoSpeedMode(  int                     nAddress, 
                                SPEED_STEP_MODE         *pSpeedStepMode );


/* ACCESSORY FUNCTIONS */

BOOL    DCPS_AccessoryDefined(  int                     nAddress );

void    DCPS_SetAccessoryState( int                     nAddress, 
                                BOOL                    bThrown );

BOOL    DCPS_AccessoryState(    int                     nAddress );

BOOL    DCPS_GetAccessoryInfo(  pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                                int                     nAddress );


/* LIST UTILITIES */


pLOCO_STRUCT
        DCPS_AddLoco(           int                     nAddress );

pLOCO_STRUCT
        DCPS_FindLoco(          int                     nAddress );

int     DCPS_GetRegisterIndex(       void );

/* S88 UTILITIES */

void    DCPS_InitS88(           void );

void    DCPS_S88UpdateRequest(  void );

BOOL    DCPS_ExtractS88State( unsigned char           *xRcvByte );

BOOL    DCPS_S88StateChanged(   void );

                                
/*======================================================================*/
/*                             STATIC VARIABLES                         */
/*======================================================================*/

static  pLOCO_STRUCT            pLocoList;

static  pLOCO_STRUCT            xLocoRegister[MAX_NB_DCCpp_LOC_REGS];

/*      accessory (turnout or signal) info:                     */
/*      - ID is address                                         */
/*      - module address is (address - 1) / 4                   */
/*      - local adress in module is (address - 1) % 4           */
/*                                                              */
/*      - state (1:Thrown 0:Unthrown) is stored in bits 0 and 1 */
/*      --- bit 0: Thrown pulse                                 */
/*      --- bit 1: Untrhown pulse                               */
/*      --- 0 0 on these 2 bits means undefined.                */
/*      - defined (stored in EEPROM is stored in bit 7          */

unsigned char                   xAccessoryState[MAX_NB_ACCESSORIES];

/*      detector info:                                          */
/*      - ID is address + 4096                                  */
/*      TBD                                                     */

/* detector info */
unsigned char                   xDetectorState[MAX_NB_DETECTORS];




/*======================================================================*/
/*                              FUNCTION DEFINITIONS                    */
/*======================================================================*/


/*======================================================================*/
/*                      LOCO RELATED UTILITIES (DCC++)                  */
/*======================================================================*/

/*======================================================================*/
/*      DCPS_EmergencyStopAllLocos                                      */
/*      Invoked on EMERGENCY STOP                                       */

BOOL
DCPS_EmergencyStopAllLocos( DCC_CMD_TAG *pCmdTag )
{
  pLOCO_STRUCT          pLoco;
  int                   nAddress;
  int                   nSpeed;
  SND_CMD_MNEMONIC      nSendCommand;
  BOOL                  bOK = TRUE;


  pLoco = pLocoList;
  while ( pLoco != NULL && bOK )
  {
    pLoco->bEmergencyStop = TRUE;

    nAddress = pLoco->nAddress;
    nSendCommand = SND_LOCO_SPEED_128;

    if ( pLoco->nRegister > 0 )
      xDataArray[0] = pLoco->nRegister; 
    else
      xDataArray[0] = 1; 
    xDataArray[1] = nAddress;
    xDataArray[2] = -1;                 // emergency stop
    xDataArray[3] = pLoco->bForward;    // forward

    /* schedule as local command : NO TAG */
    bOK = DCPT_ScheduleSndMsg( hMainWnd, NULL, nSendCommand, 4, 
                               xDataArray, FALSE, CMD_LOCAL, PRI_2 );

    pLoco = pLoco->pNext;
  }

  return bOK;
}

/*======================================================================*/
/*      DCPS_RestartAllLocos                                            */
/*      INVOKED ON RESUME OPERATIONS if in EMERGENCY STOP state.        */
/*      Scans the loco list, and restores all loco speeds, before       */
/*      Emergency stop.                                                 */
/*      Resets the bEmergencyStop field.                                */

BOOL
DCPS_RestartAllLocos( void )
{
  pLOCO_STRUCT          pLoco;
  int                   nAddress;
  int                   nSpeed;
  SND_CMD_MNEMONIC      nSendCommand;
  BOOL                  bOK = TRUE;

  pLoco = pLocoList;
  while ( pLoco != NULL && bOK )
  {
    if ( pLoco->bEmergencyStop )
    {
      pLoco->bEmergencyStop = FALSE;

      nAddress = pLoco->nAddress;
      nSendCommand = SND_LOCO_SPEED_128;

      if ( pLoco->nRegister > 0 )
        xDataArray[0] = pLoco->nRegister;
      else
        xDataArray[0] = pLoco->nRegister = DCPS_GetRegisterIndex() + 1;
      xDataArray[1] = nAddress;
      xDataArray[2] = pLoco->nSpeed;      // restore speed
      xDataArray[3] = pLoco->bForward;    // forward

      /* schedule as local command */
      bOK = DCPT_ScheduleSndMsg( hMainWnd, NULL, nSendCommand, 4, 
                                 xDataArray, FALSE, CMD_LOCAL, PRI_2 );

    }   // end if pLoco->bEmergencyStop )
    pLoco = pLoco->pNext;
  }

  return bOK;
}


/*======================================================================*/
/*                             LOCO UTILITIES                           */
/*======================================================================*/



/*======================================================================*/
/*      DCPS_SetFxByte                                                  */
/*      Sets *pDataByte to a 4 or 5 bit value according to the value    */
/*      nFuncMask.                                                      */
/*      0x001F:         128 (0x80)+ bits 4 to 0. swap bit 0 to bit4.    */
/*      0x01E0:         176 (0xB0) + bits  8 to 5.                      */
/*      0x1E00:         160 (0xA0) + bits 12 to 9.                      */
/*                                                                      */

unsigned char
DCPS_SetFxByte( unsigned int    nFuncValue, 
                unsigned int    nFuncMask )
{
  unsigned char         cResult;
  unsigned int          nValue;
  BOOL                  bF0;

  switch( nFuncMask )
  {
  case (unsigned int) 0x001F:
    nValue = nFuncValue & nFuncMask;
    bF0 = (nValue & 1) != 0;
    nValue >>= 1;
    if ( bF0 )
      nValue |= (unsigned int) 0x10;
    nValue += (unsigned int) 0x80;
    break;

  case (unsigned int) 0x01E0:
    nValue = nFuncValue & nFuncMask;
    nValue >>= 5;
    nValue += (unsigned int) 0xB0;
    break;

  case (unsigned int) 0x1E00:
    nValue = nFuncValue & nFuncMask;
    nValue >>= 9;
    nValue += (unsigned int) 0xA0;
    break;

  default:
    break;

  }

  return ((unsigned char) nValue);
}


/*======================================================================*/
/*      DCPS_UpdateLocoParameters                                       */
/*      This function:                                                  */
/*      - seraches for loco at address nAddress, in pLocoList.          */
/*      - if not found, adds a new LOCO_STRUCT to that list.            */
/*      - then sets all fields of LOCO STRUCT to specified parameters.  */
/*      - returns the DCCpp REGISTER number allocated to that loco.     */

int   
DCPS_UpdateLocoParameters( int                  nAddress, 
                        int                     nSpeed, 
                        SPEED_STEP_MODE         nSpeedStepMode, 
                        BOOL                    bForward,
                        unsigned int            nF0F12,
                        unsigned int            nF0F12Mask )
{
  pLOCO_STRUCT          pLoco;

  pLoco = DCPS_FindLoco( nAddress );
  if ( pLoco !=  NULL )
  {
    // if no register allocated, reallocate.
    if ( pLoco->nRegister == 0 )
    {
      pLoco->nRegister = DCPS_GetRegisterIndex() + 1;        // start from nb 1
      xLocoRegister[pLoco->nRegister - 1] = pLoco;
    }
  }
  else
    pLoco = DCPS_AddLoco( nAddress );

  if ( pLoco == NULL )
     MessageBox( hMainWnd, "FATAL ERROR: COULD NOT CREATE LOCO STRUCT", 
                 "LOCO PARAMETERS", 
                 MB_APPLMODAL | MB_OK | MB_ICONERROR );
  else
  {
    if ( !xDccCtxt.xServerDesc.bEmergencyStop )
      pLoco->bEmergencyStop = FALSE;
    pLoco->nAddress = nAddress;
    pLoco->nSpeed = nSpeed;
    pLoco->nMode = nSpeedStepMode;
    pLoco->bForward = bForward;
    pLoco->nF0F12 &= ~nF0F12Mask;
    pLoco->nF0F12 |= ( nF0F12 & nF0F12Mask );
    pLoco->lLastTimeTick = xDccCtxt.xServerDesc.lCurTimeTick;
  }

  return pLoco->nRegister;
}

/*======================================================================*/
/*      DCPS_UpdateLocoFx                                               */
/*      Idem DCPU_UpdateLocoParameters, for functions only.             */

void    
DCPS_UpdateLocoFx( int nAddress, unsigned int nF0F12, unsigned int nF0F12Mask )
{
  pLOCO_STRUCT          pLoco;

  pLoco = DCPS_FindLoco( nAddress );
  if ( pLoco == NULL )
    pLoco = DCPS_AddLoco( nAddress );

  if ( pLoco == NULL )
     MessageBox( hMainWnd, "FATAL ERROR: COULD NOT CREATE LOCO STRUCT", 
                 "LOCO PARAMETERS", 
                 MB_APPLMODAL | MB_OK | MB_ICONERROR );
  else
  {
    pLoco->nF0F12 &= ~nF0F12Mask;
    pLoco->nF0F12 |= nF0F12;
  }
    
}

/*======================================================================*/
/*      DCPS_SetEmergencyStop                                           */
/*      This function:                                                  */
/*      - seraches for loco at address nAddress, in pLocoList.          */
/*      - if not found, adds a new LOCO_STRUCT to that list.            */
/*      - then sets all fields of LOCO STRUCT to specified parameters.  */

void    
DCPS_SetEmergencyStop( int nAddress )
{
  pLOCO_STRUCT          pLoco;

  pLoco = DCPS_FindLoco( nAddress );
  if ( pLoco == NULL )
    pLoco = DCPS_AddLoco( nAddress );

  if ( pLoco == NULL )
     MessageBox( hMainWnd, "FATAL ERROR: COULD NOT CREATE LOCO STRUCT", 
                 "LOCO PARAMETERS", 
                 MB_APPLMODAL | MB_OK | MB_ICONERROR );
  else
    pLoco->bEmergencyStop = TRUE;
    
}

/*======================================================================*/
/*      DCPS_GetLocoInfo                                                */
/*                                                                      */
/*      - pGenericRcvData->xDataItem[0].hData[0] : speed                */
/*      - pGenericRcvData->xDataItem[0].hData[1] : F0F12                */
/*      - pGenericRcvData->xDataItem[0].hData[2] : function mask        */
/*                              (bits in hdata[1] relevant or not)      */
/*      - pGenericRcvData->xDataItem[0].hData[3] : status.              */
/*                                                      0 in that case. */

BOOL   
DCPS_GetLocoInfo( pDGI_GENERIC_RCV_DATA pGenericRcvData, int nAddress )
{
  pLOCO_STRUCT          pLoco;
  BOOL                  bOK = TRUE;

  pLoco = DCPS_FindLoco( nAddress );
  if ( pLoco == NULL )
    bOK = FALSE;

  else
  {
    pGenericRcvData->xDataItem[0].hData[0] = (short) pLoco->nSpeed;
    pGenericRcvData->xDataItem[0].hData[1] = (short) pLoco->nF0F12;
    pGenericRcvData->xDataItem[0].hData[2] = (short) 0x01FFF;
    pGenericRcvData->xDataItem[0].hData[3] = 0;
  }

  return bOK;
}

/*======================================================================*/
/*      DCPS_GetLocoSpeedMode                                           */
/*      returns TRUE if loco fund, with loco speed step mode in         */
/*      *pSpeedStepMode, or FALSE if loco not found.                    */

BOOL
DCPS_GetLocoSpeedMode( int nAddress, SPEED_STEP_MODE *pSpeedStepMode )
{
  pLOCO_STRUCT          pLoco;
  BOOL                  bOK = TRUE;

  pLoco = DCPS_FindLoco( nAddress );
  if ( pLoco == NULL )
    bOK = FALSE;
  else 
    *pSpeedStepMode = pLoco->nMode;

  return bOK;
}

/*======================================================================*/
/*                         ACCESSORY FUNCTIONS                          */
/*======================================================================*/


/*======================================================================*/
/*      DCPS_AccessoryDefined                                           */
/*      Stored or not in DCC++ EEPROM                                   */

BOOL
DCPS_AccessoryDefined( int nAddress )
{
  if ( ( xAccessoryState[nAddress] & ((unsigned char) ACC_THROWN_MASK ) ) != 0 || 
       ( xAccessoryState[nAddress] & ((unsigned char) ACC_UNTHROWN_MASK ) ) != 0 )
    return TRUE;
  else
    return FALSE;
}

/*======================================================================*/
/*      DCPS_SetAccessoryState                                          */

void
DCPS_SetAccessoryState( int nAddress, BOOL bThrown )
{
  if ( bThrown )
  {
    xAccessoryState[nAddress] |= (unsigned char) ACC_THROWN_MASK;
    xAccessoryState[nAddress] &= (unsigned char) ~ACC_UNTHROWN_MASK;
  }
  else
  {
    xAccessoryState[nAddress] &= (unsigned char) ~ACC_THROWN_MASK;
    xAccessoryState[nAddress] |= (unsigned char) ACC_UNTHROWN_MASK;
  }
}

/*======================================================================*/
/*      DCPS_AccessoryState                                             */
/*      Stored or not in DCC++ EEPROM                                   */
/*      RETURN: Thrown (TRUE) or UNTHROWN (FALSE)                       */

BOOL
DCPS_AccessoryState( int nAddress )
{
  if ( xAccessoryState[nAddress] & ((unsigned char) ACC_THROWN_MASK ) != 0 )
    return TRUE;
  else
    return FALSE;
}

/*======================================================================*/
/*      DCPS_GetAccessoryInfo                                           */
/*      Invoked for returning info to CDM-Rail.                         */
/*      returns the 2 bits info on THROWN/UNTHROWN pulses.              */

BOOL
DCPS_GetAccessoryInfo(  pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                        int                     nAddress )
{
  unsigned char         cPulseValues;

  cPulseValues = xAccessoryState[nAddress] &
                ( ACC_THROWN_MASK | ACC_UNTHROWN_MASK );
  pGenericRcvData->xDataItem[0].hData[0] = (unsigned short) cPulseValues;
}


/*======================================================================*/
/*                             SPEED UTILITIES                          */
/*======================================================================*/

/*======================================================================*/
/*      DCPS_AddLoco                                                    */
/*      Adds a loco strcuture in list pLocoList, in address order.      */


pLOCO_STRUCT
DCPS_AddLoco( int nAddress )
{
  pLOCO_STRUCT          pLoco, pLoco2, pPrevLoco;
  BOOL                  bFound;
  int                   nRegister;

  // fisrt retrieve register for new loc. BEFORE creating loco
  nRegister = DCPS_GetRegisterIndex() + 1;

  pLoco = calloc( 1, sizeof( struct S_LOCO_STRUCT ) );
  if ( pLoco == NULL ) return NULL;

  if ( pLocoList == NULL )
    pLocoList = pLoco;
  else
  {
    pLoco2 = pLocoList;
    if ( nAddress < pLoco2->nAddress )
    {
      /* insert in first place */
      pLocoList = pLoco;
      pLoco2->pPrev = pLoco;
      pLoco->pNext = pLoco2;
    }  
    else
    {
      bFound = FALSE;
      while ( !bFound && pLoco2->pNext!= NULL )
      {
        pLoco2 = pLoco2->pNext;
        if ( nAddress < pLoco2->nAddress )
          bFound = TRUE;
      }         /* end while */

      if ( !bFound )    /* insert at end of list, after pLoco2 */
      {
        pLoco2->pNext = pLoco;
        pLoco->pPrev = pLoco2;
      }
      else              /* insert before pLoco2 */
      {
        pPrevLoco = pLoco2->pPrev;
        pLoco->pNext = pLoco2;
        pLoco2->pPrev = pLoco;
        pLoco->pPrev = pPrevLoco;
        if ( pPrevLoco != NULL )
          pPrevLoco->pNext = pLoco;  
      }
    }
  }

  if ( pLoco != NULL )
  {
    pLoco->nAddress = nAddress;
    pLoco->nRegister = nRegister;
    xLocoRegister[pLoco->nRegister - 1] = pLoco;
  }


  return pLoco;

}

/*======================================================================*/
/*      DCPS_FindLoco                                                   */
/*      Seraches for a loco strcuture in list pLocoList, with address   */
/*      nAddress.                                                       */


pLOCO_STRUCT
DCPS_FindLoco( int nAddress )
{
  pLOCO_STRUCT          pLoco, pLoco2, pPrevLoco;
  BOOL                  bFound;

  bFound = FALSE;

  pLoco = pLocoList;
  while ( pLoco != NULL && !bFound )
  {
    if ( pLoco->nAddress == nAddress )
      bFound = TRUE;
    else
      pLoco = pLoco->pNext;
  }

  return pLoco;

}

/*======================================================================*/
/*      DCPS_GetRegisterIndex                                           */
/*      Gets a free register, to be allocated to a new loco.            */
/*      If no register is available, DCPS_RequestRegister is invoked    */
/*      for deallocating a busy register from the corresponding loco.   */
/*      NOTE:                                                           */
/*      The "register index" ranges from 0 to MAX_NB_DCCpp_LOC_REGS - 1 */
/*      whereas the corresponding register number ranges from 1 to      */
/*      MAX_NB_DCCpp_LOC_REGS.                                          */

int
DCPS_GetRegisterIndex( void )
{
  int           nRegisterIndex;
  BOOL          bFound = FALSE;
  
  nRegisterIndex = 0;
  while( !bFound && nRegisterIndex < MAX_NB_DCCpp_LOC_REGS )
  {
    if ( xLocoRegister[nRegisterIndex] == NULL )
      bFound = TRUE;
    else
      nRegisterIndex++;
  }

  if ( !bFound )
  {
    nRegisterIndex = DCPS_RequestRegisterIndex();
  }

  return nRegisterIndex;
}

/*======================================================================*/
/*      DCPS_RequestRegisterIndex                                       */
/*      Invoked when all 12 registers are allocated to locos.           */
/*      This routine looks for a loco with 0 speed, and earliest time   */
/*      tick. The register of this loco is stolen for new use.          */

int
DCPS_RequestRegisterIndex( void )
{
  pLOCO_STRUCT          pLoco, pSelectedLoco;
  long                  lLowestTimeTick;
  int                   nRegisterIndex;
  int                   nIndex;
  char                  sMessage[64];

  nIndex = 0;
  pSelectedLoco = xLocoRegister[nIndex];
  lLowestTimeTick = pSelectedLoco->lLastTimeTick;

  for ( nIndex = 1; nIndex < MAX_NB_DCCpp_LOC_REGS; nIndex++ )
  {
    pLoco = xLocoRegister[nIndex];

    if ( pLoco->nSpeed == 0 )    
    {
      if ( pSelectedLoco->nSpeed != 0 )
      {
        pSelectedLoco = pLoco;
        lLowestTimeTick = pLoco->lLastTimeTick;        
      }
      else if ( pLoco->lLastTimeTick < lLowestTimeTick )
      {
        pSelectedLoco = pLoco;
        lLowestTimeTick = pLoco->lLastTimeTick;
      }
    }

    else        // speed !0
    {
      if ( pSelectedLoco->nSpeed == 0 )
      { /* do not change */ }
      
      // both locs running: take oldest time tick.
      else if ( pLoco->lLastTimeTick < lLowestTimeTick )
      {
        pSelectedLoco = pLoco;
        lLowestTimeTick = pLoco->lLastTimeTick;
      }    
    }
  }

  nRegisterIndex = pSelectedLoco->nRegister - 1;
  pSelectedLoco->nRegister = 0;
  xLocoRegister[nRegisterIndex] = NULL;
  
  return nRegisterIndex;
}

/*======================================================================*/
/*      DCPS_InitS88                                                    */
/*      Send command "Y 16 1>" for S88 init.                            */
/*      16 is the number of 8 modules: 128 detectors.                   */
/*      1 stands for hexa format.                                        */

void
DCPS_InitS88( void )
{
  BOOL          bOK;

  xDataArray[0] = MAX_NB_DETECTORS / 8;         // nb of 8_det_modules
  xDataArray[1] = 1;  // hexadecimal

  /* schedule as local command : NO TAG */
  bOK = DCPT_ScheduleSndMsg( hMainWnd, NULL, SND_INIT_S88, 2, 
                               xDataArray, FALSE, CMD_LOCAL, PRI_0 );
}

/*======================================================================*/
/*      DCPS_S88UpdateRequest                                            */
/*      Send command "Y>" for S88 update.                               */

void
DCPS_S88UpdateRequest( void )
{
  BOOL          bOK;

  /* schedule as local command : NO TAG */
  bOK = DCPT_ScheduleSndMsg( hMainWnd, NULL, SND_S88_UPDATE_REQUEST, 0, 
                               xDataArray, FALSE, CMD_LOCAL, PRI_0 );
}

/*======================================================================*/
/*      DCPS_ExtractS88State                                            */
/*      Scans all 32 hexadecimal character of received message. Any     */
/*      character is the most recent state of a Nibble (4 bits).        */
/*      The function stores the new state in array xS88State: one entry */
/*      par nibble:                                                     */
/*      - bit 3-0;      current state                                   */
/*      - bit 7-4:      prev state                                      */
/*      For any nibble changed, a bit is set in nS88NibbleChanges       */
     
BOOL
DCPS_ExtractS88State( unsigned char *xRcvByte )
{
  char                  sHeader[MAX_STR_LEN], sS88String[MAX_STR_LEN];
  BOOL                  bOK = TRUE;
  int                   nNibble;
  int                   nNewValue, nCurValue;
  int                   nChange;
  int                   i;
  char                  sMessage[128];

  bS88StateUpdated = TRUE;

  // fisrt trim last '>' 
  i = strlen( xRcvByte );
  while ( i >0 && xRcvByte[i] != '>' ) i--;

  if ( xRcvByte[i] != '>' )
    bOK = FALSE;
  else
    xRcvByte[i] = '\0';

  if ( sscanf( xRcvByte, "%s %s" , sHeader, sS88String ) != 2 )
  {
    if ( pDebugFile != NULL )
    {
      sprintf( sMessage , "DCPS_ExtractS88State err 1: %s", xRcvByte );
      DCPU_PrintDebugMessage( sMessage );
    }
    bOK = FALSE;
  }
  else if ( strlen( sS88String ) != ( MAX_NB_DETECTORS / 4 ) )
  {
    if ( pDebugFile != NULL )
    {
      sprintf( sMessage , "DCPS_ExtractS88State err 2 : %d / %%d",
                strlen( sS88String ), ( MAX_NB_DETECTORS / 4 ) );
      DCPU_PrintDebugMessage( sMessage );
    }
    bOK = FALSE;
  }
  else
  {
    if ( pDebugFile != NULL )
    {
      sprintf( sMessage , "DCPS_ExtractS88State OK" );
      DCPU_PrintDebugMessage( sMessage );
    }

    for ( nNibble = 0; nNibble < (MAX_NB_DETECTORS / 4) && bOK ; nNibble++ )
    {
      bOK = DCPU_Ascii2Hexa( &sS88String[nNibble], 1, &nNewValue );
      if ( bOK )
      {
        nCurValue = xS88State[nNibble] & ((unsigned char) (0x0F));
        if ( nCurValue != nNewValue )
        {
          if ( pDebugFile != NULL )
          {
            sprintf( sMessage , "nibble #%d old value %02lX new value %02lX",
                nNibble, nCurValue, nNewValue );
            DCPU_PrintDebugMessage( sMessage );
          }

          // shift current value to old value, and set new value.
          xS88State[nNibble] <<= 4;
          xS88State[nNibble] |= nNewValue;
          nChange = 1 << nNibble;
          nS88NibbleChanges |= nChange;
        }
      }
    }
  }

  return bOK;
}

/*======================================================================*/
/*      DCPS_S88StateChanged()                                          */

BOOL
DCPS_S88StateChanged( void )
{
  return (nS88NibbleChanges != 0);
}
