/*======================================================================*/
/*                                                                      */
/*      dccpp_thread.c                                                  */
/*      This module contains all functions pertaining to read/write     */
/*      oper. (send message to interface).                              */
/*======================================================================*/
/*      J.P.PILLOU      13/03/2018                                      */
/*======================================================================*/
/*======================================================================*/

#include <windows.h>
#include <stdio.h>
#include "DDGI_Generic_def.h"
#include "dccpp_def.h"
#include "dccpp_extern.h"
#include "dccpp_thread.h"
#include "dccpp_specific.h"
#include "dccpp_util.h"

/*======================================================================*/
/*                         FUNCTION DECLARATIONS                        */
/*======================================================================*/

/* functions in main process space */
/*=================================*/

BOOL    DCPT_ScheduleSndMsg(    HWND                    hWnd,
                                DCC_CMD_TAG            *pCmdTag,
                                SND_CMD_MNEMONIC        nCommand, 
                                int                     nNbArgs,
                                int                     pDataArray[],
                                BOOL                    bShowCommand,
                                CMD_TYPE                nCommandType,
                                CMD_PRIORITY            nPriority );

BOOL    DCPT_GetNextExpectedResponse( 
                        pDGI_GENERIC_RCV_DATA   pGenericRcvData );

BOOL    DCPT_GetNextUnExpectedMessage( 
                        pDGI_GENERIC_RCV_DATA   pGenericRcvData );

/* functions in write thread space */
/*=================================*/

DWORD   DCPT_IFThreadFunc(      void                    *Ptr );

BOOL    DCPT_SendCommand(       HWND                    hWnd, 
                                HANDLE                  hCom, 
                                DCC_COMMAND_CTXT        *pSndCommandCtxt,
                                BOOL                    bShow );

RD_ERROR 
        DCPT_ReadCommand(       HWND                    hWnd,
                                HANDLE                  hCom,
                                DCC_COMMAND             *pRcvCommand,
                                unsigned char           xRcvBuffer[],
                                int                     nNbMaxLatency,
                                RCV_CMD_MNEMONIC        nExpectedRdCommand,
                                unsigned char           *sExpectedRespHeader,
                                BOOL                    bExpected,
                                BOOL                    bShowCommand,
                                BOOL                    bStoreToFifo );

RD_ERROR
        DCPT_LookForErrorCode(  unsigned char           *xRcvByte, 
                                int                     nNbBytes );

RD_ERROR
        DCPT_LookForUnexpectedResponse(  
                                DCC_COMMAND             *pRcvCommand,
                                unsigned char           *sExpectedRespHeader,
                                unsigned char           *xRcvByte, 
                                int                     nNbBytes );

RD_ERROR
        DCPT_GetResponse(       DCC_COMMAND             *pRcvCommand,
                                RCV_CMD_MNEMONIC        nExpectedResponse,
                                unsigned char           *xRcvByte, 
                                int                     nNbBytes );

BOOL    DCPT_ScanForResponse(   HWND                    hWnd, 
                                HANDLE                  hCom, 
                                DCC_COMMAND_CTXT        *pSndCommandCtxt,
                                BOOL                    bShow );

BOOL    DCPT_PerformStandaloneHandling( HWND            hWnd, 
                                DCC_COMMAND_CTXT        *pSndCommandCtxt, 
                                BOOL                    bShow);

void    DCPT_PerformExtraHandling( HWND                 hWnd, 
                                DCC_COMMAND_CTXT        *pSndCommandCtxt, 
                                BOOL                    bShow);

BOOL    DCPT_GenerateEventMessage( RCV_CMD_MNEMONIC     nRdCommand );

BOOL    DCPT_LookForUnexpectedMessage(   
                                HWND                    hWnd, 
                                HANDLE                  hCom, 
                                DCC_COMMAND_CTXT        *pRcvCommandCtxt,
                                BOOL                    bShow );

BOOL    DCPT_HandleUnexpectedMessage( 
                                DCC_COMMAND_CTXT        *pCommandCtxt,
                                RCV_CMD_MNEMONIC        nRdCommand,
                                unsigned char           xRcvByte[],
                                BOOL                    bOutCommandCycle );

void    DCPT_DisplayErrorMessage( HWND                  hWnd,
                                DCC_COMMAND_CTXT        *pCommandCtxt );


/*======================================================================*/
/*                     MAIN PROCESS FUNCTION DEFINITIONS                */
/*======================================================================*/

/*======================================================================*/
/*      DCPT_ScheduleSndMsg                                             */
/*      This function shedules a send message.                          */
/*      It creates (or gets from free element list) a new DCC_CMD_CTXT  */
/*      and appends it to the list of send commands:                    */
/*      xDccCtxt.pSndCmdCtxtList.                                       */
/*      If bUrgent is TRUE, then the new command is inserted immediately*/
/*      after the command being processed.                              */
/*      If the number of elements exceeds MAX_SND_MSG, then the number  */
/*      of overflows is incremented, and the function returns FALSE.    */
/*      If a new command is redundant with another command already      */
/*      stored and not yet executed, the new command is NOT stored.     */

BOOL
DCPT_ScheduleSndMsg(    HWND                    hWnd,
                        DCC_CMD_TAG            *pCmdTag,
                        SND_CMD_MNEMONIC        nCommand,
                        int                     nNbArgs, 
                        int                     *pDataArray, 
                        BOOL                    bShowCommand,
                        CMD_TYPE                nCommandType,
                        CMD_PRIORITY            nPriority )
{
  DWORD                 lCount;
  pDCC_COMMAND_CTXT     pCmdCtxt;
  BOOL                  bOK = TRUE;
  char                  sMessage[128];

  EnterCriticalSection( &xDccCtxt.xCriticalSection );

  if ( xDccCtxt.xSndCmd.nNbCmd >= MAX_SND_MSG && nPriority < PRI_3 )
  {
    xDccCtxt.lNbSndOverflows++;
    bOK = FALSE;
  }
  else
  {
    pCmdCtxt = DCPU_NewCommandCtxt( &xDccCtxt.xSndCmd, nPriority );
    xDccCtxt.lTotalNbCmdSent++;
    if ( nCommandType == CMD_LOCAL )
      xDccCtxt.lTotalNbLocalCmdSent++;
    else
      xDccCtxt.lTotalNbUserCmdSent++;

    pCmdCtxt->nSndSeqState = DCC_SEQ_SCHEDULED;
    pCmdCtxt->nSndCommand = nCommand;
    pCmdCtxt->nCommandType = nCommandType;
    DCPU_CopySndCommand( nCommand, nNbArgs, pDataArray, &pCmdCtxt->xSndCommand );
    pCmdCtxt->bDisplay = bShowCommand;
    
    if ( pCmdTag != NULL )
      pCmdCtxt->xCmdTag = *pCmdTag;
    else
      DCPU_ClearCmdTag( &pCmdCtxt->xCmdTag );
  }

  LeaveCriticalSection( &xDccCtxt.xCriticalSection );

  return bOK;
}

/*======================================================================*/
/*      DCPT_GetNextExpectedResponse                                    */
/*      Detaches first Cmd Ctxt of list xDccCtxt.xSndCmd, (which is     */
/*      thus added to free element list xDccCtxt.pFreeCmdCtxtList.      */
   
BOOL    
DCPT_GetNextExpectedResponse( pDGI_GENERIC_RCV_DATA pGenericRcvData ) 
{
  BOOL                  bFound = FALSE;
  pDCC_COMMAND_CTXT     pCmdCtxt, pErrorCtxt;
  RCV_CMD_MNEMONIC      nCommand;
  int                   nAddress;
  int                   xData[MAX_NB_DCC_DATA_WORDS];
  int                   nNbBytes;
  BOOL                  bEnd;
  unsigned char         xByte[MAX_NB_CMD_DATA];

  char                  sMessage[64];

  EnterCriticalSection( &xDccCtxt.xCriticalSection );

  bEnd = FALSE;
  while ( !bEnd )
  {
    if ( xDccCtxt.xSndCmd.pCmdCtxtList != NULL )
    {
      pCmdCtxt = xDccCtxt.xSndCmd.pCmdCtxtList;
      if ( pCmdCtxt->nSndSeqState == DCC_SEQ_DONE )
      /* if ( xDccCtxt.pSndCmdRdPtr != NULL &&
           xDccCtxt.pSndCmdRdPtr != xDccCtxt.pSndCmdCtxtList ) */
      {
        if ( xDccCtxt.xSndCmd.pCmdRdPtr == xDccCtxt.xSndCmd.pCmdCtxtList )
          xDccCtxt.xSndCmd.pCmdRdPtr = NULL;
        bFound = TRUE;
        pGenericRcvData->xCmdTag = pCmdCtxt->xCmdTag;
        pGenericRcvData->nCompletionCode = pCmdCtxt->nCompletionCode;

        if ( pDebugFile != NULL )
        {
          sprintf( sMessage , 
                        "Get cmd %d (%ld) expected response %d completion %d", 
                        pCmdCtxt->nSndCommand,
                        pCmdCtxt->lIndex,
                        pCmdCtxt->nRdCommand,
                        pCmdCtxt->nCompletionCode );
          DCPU_PrintDebugMessage( sMessage );
        }

        /* if error, store command to xError environment */
        if ( pCmdCtxt->nCompletionCode != DGI_COMP_OK )
        {
          if ( pDebugFile != NULL )
          {
            sprintf( sMessage , "Get cmd %d (%ld) expected response: ERROR",
                                     pCmdCtxt->nSndCommand, pCmdCtxt->lIndex );
            DCPU_PrintDebugMessage( sMessage );
          }

           xDccCtxt.lTotalNbErrors++;
          if ( xDccCtxt.xError.nNbCmd >= MAX_ERR_MSG )
          {
            xDccCtxt.lNbErrOverflows++;
            DCPU_DetachCommandCtxt( &xDccCtxt.xError );
          }
          pErrorCtxt = DCPU_NewCommandCtxt( &xDccCtxt.xError, PRI_0 );
          DCPU_CopyCommandCtxt( pErrorCtxt, pCmdCtxt );
        }

        else      /* DGI_COMP_OK: check if response is different from RCV_ACK */
        {
        
          if ( pCmdCtxt->nRdCommand != RCV_ACK )
          {  
            pGenericRcvData->nCompletionCode = 
                pCmdCtxt->nCompletionCode = DGI_COMP_OK_RD; 
            xDccCtxt.xCurCmdResponse = *pCmdCtxt;

            nCommand = pCmdCtxt->nRdCommand;
            DCPU_MapResponseToGeneric( pGenericRcvData, pCmdCtxt, &nCommand );

            pGenericRcvData->nRdCommand = nCommand;
            if ( pDebugFile != NULL )
            {
              sprintf( sMessage , "Get cmd %d (%ld) Rdcommand %d Completion %d",
                                pCmdCtxt->nSndCommand, pCmdCtxt->lIndex,
                                pGenericRcvData->nRdCommand,
                                pGenericRcvData->nCompletionCode  );
              DCPU_PrintDebugMessage( sMessage );
            }
          }
        }       /* end DGI_COMP_OK */
          
        if ( pCmdCtxt->nCommandType != CMD_LOCAL &&
             pCmdCtxt->nCommandType != CMD_REFRESH )
          bEnd = TRUE;
        else            /* get next */
          bEnd = FALSE;
        DCPU_DetachCommandCtxt( &xDccCtxt.xSndCmd );
      }         /* end if if ( pCmdCtxt->nSndSeqState == DCC_SEQ_DONE ) */

      else      /* not DONE */
        bEnd = TRUE;
    }           /* end if if ( xDccCtxt.xSndCmd.pCmdCtxtList != NULL ) */

    else        /* xDccCtxt.xSndCmd.pCmdCtxtList == NULL */
      bEnd = TRUE;

  }     /* end while */
  LeaveCriticalSection( &xDccCtxt.xCriticalSection );

  return bFound;
}

/*======================================================================*/
/*      DCPT_GetNextUnExpectedMessage                                   */
/*      Detaches first Cmd Ctxt of list xDccCtxt.xRcvMsg, (which is     */
/*      thus added to free element list xDccCtxt.pFreeCmdCtxtList.      */
   
BOOL    
DCPT_GetNextUnExpectedMessage( pDGI_GENERIC_RCV_DATA pGenericRcvData )
{
  BOOL                  bFound = FALSE;
  RCV_CMD_MNEMONIC      nCommand;
  int                   nAddress;
  int                   xData[MAX_NB_DCC_DATA_WORDS];
  int                   nNbBytes; 
  unsigned char         xByte[MAX_NB_CMD_DATA];  
  pDCC_COMMAND_CTXT     pCmdCtxt, pErrorCtxt;

  EnterCriticalSection( &xDccCtxt.xCriticalSection );
  if ( xDccCtxt.xRcvMsg.pCmdCtxtList != NULL )
  {
    pCmdCtxt = xDccCtxt.xRcvMsg.pCmdCtxtList;
    bFound = TRUE;

    /* if error, store command to xError environment */
    if ( pCmdCtxt->nCompletionCode != DGI_COMP_OK )
    {
      xDccCtxt.lTotalNbErrors++;
      if ( xDccCtxt.xError.nNbCmd >= MAX_ERR_MSG )
      {
        xDccCtxt.lNbErrOverflows++;
        DCPU_DetachCommandCtxt( &xDccCtxt.xError );
      }
      pErrorCtxt = DCPU_NewCommandCtxt( &xDccCtxt.xError, PRI_0 );
      DCPU_CopyCommandCtxt( pErrorCtxt, pCmdCtxt );
    }
    else
    {
      pGenericRcvData->nCompletionCode = DGI_COMP_OK;
      pGenericRcvData->nRdCommand = nCommand = pCmdCtxt->nRdCommand;
      DCPU_MapMessageToGeneric( pGenericRcvData, pCmdCtxt, &nCommand );
    }

    DCPU_DetachCommandCtxt( &xDccCtxt.xRcvMsg );
  }
  LeaveCriticalSection( &xDccCtxt.xCriticalSection );

  return bFound;
}


/*======================================================================*/
/*                     THREAD SPACE FUNCTION DEFINITIONS                */
/*======================================================================*/

/*======================================================================*/
/*      DCPT_IFThreadFunc                                               */
/*      This is the entry point function of the write thread.           */
/*      The function consists of the following infinite loop.           */
/*                                                                      */
/*      - If there is a schedule SEND command:                          */
/*        * Send this command.  (DCCT_SendCommand)                      */
/*        * Wait for response.  ( DCCT_ScanForResponse).                */
/*        * If response id CMD station busy: re-send command.           */
     
DWORD
DCPT_IFThreadFunc( void *Ptr )
{
  DWORD                 lReturn = 0;
  HWND                  hWnd;
  BOOL                  bDisplayError = FALSE;
  int                   nNewSndMsgRdPrt;
  int                   nSndMsgPtr;
  DCC_COMMAND_CTXT      xRcvCommandCtxt;
  pDCC_COMMAND_CTXT     pCmdCtxt;
  BOOL                  bDone;
  char                  sMessage[MAX_LONG_MSG_LEN];
  static long           lTimeTick = 0L;

  hWnd = (HWND) Ptr;

  while ( 1 )
  {
    if ( xDccCtxt.xSndCmd.nNbPendingCmd > 0 )
    {
      if ( pDebugFile != NULL )
      {
        sprintf( sMessage , "Nb pending cmds: %d", xDccCtxt.xSndCmd.nNbPendingCmd );
        DCPU_PrintDebugMessage( sMessage );
      }

      EnterCriticalSection( &xDccCtxt.xCriticalSection );

      if ( xDccCtxt.xSndCmd.pCmdRdPtr == NULL )
        xDccCtxt.xSndCmd.pCmdRdPtr = xDccCtxt.xSndCmd.pCmdCtxtList;
      else
        xDccCtxt.xSndCmd.pCmdRdPtr = xDccCtxt.xSndCmd.pCmdRdPtr->pNext;
      
      pCmdCtxt = xDccCtxt.xSndCmd.pCmdRdPtr;
      pCmdCtxt->nSndSeqState = DCC_SEQ_INIT;

      LeaveCriticalSection( &xDccCtxt.xCriticalSection );

      if ( pCmdCtxt->nCommandType == CMD_STANDALONE )
      {
        DCPT_PerformStandaloneHandling( hWnd, pCmdCtxt, pCmdCtxt->bDisplay );
        DCPT_PerformExtraHandling(hWnd, pCmdCtxt, pCmdCtxt->bDisplay );
      }

      else                              //normal processing
      {
        bDone = FALSE;

        while( !bDone )
        {
          if ( !DCPT_SendCommand( hWnd, xDccCtxt.hComPort,
                  pCmdCtxt, pCmdCtxt->bDisplay ) )
          {
            /*
            MessageBox( hWnd, "SEND COMMAND FAILED!!", 
                        "WRITE THREAD", 
                        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
            */
            /*
            if ( pDebugFile != NULL )
            {
              sprintf( sMessage , "send error cmd %d ", pCmdCtxt->nSndCommand );
              DCPU_PrintDebugMessage( sMessage );
            }
            */
          }
          else
          {
            if ( !DCPT_ScanForResponse( hWnd, xDccCtxt.hComPort,
                  pCmdCtxt, pCmdCtxt->bDisplay ) )
            {
              if ( pDebugFile != NULL )
              {
                sprintf( sMessage , "cmd %d %08X (%ld) response problem", 
                        pCmdCtxt->nSndCommand, 
                        (unsigned long) pCmdCtxt,
                        pCmdCtxt->lIndex  );
                DCPU_PrintDebugMessage( sMessage );
              }

              if ( DCPU_CommandStationBusy( pCmdCtxt->nRdErrorCode ) )
              {
                bDone = FALSE;            /* retry */
                xDccCtxt.lTotalNbCmdBusyMsg++;
              }
              else
              {
                DCPT_DisplayErrorMessage( hWnd, pCmdCtxt );
                DCPU_DumpDbgFifoToFile( pDumpFile );
                DCPU_DumpCmdFifoToFile( pDumpFile );
                bDone = TRUE;
                pCmdCtxt->nSndSeqState = DCC_SEQ_SENT;
              }
            }       /* if ( !DCPT_ScanForResponse */  
            else
            {
              if ( pDebugFile != NULL )
              {
                sprintf( sMessage , 
                        "cmd %d (%ld) response received OK completion %d", 
                        pCmdCtxt->nSndCommand, 
                        pCmdCtxt->lIndex, 
                        pCmdCtxt->nCompletionCode );
                DCPU_PrintDebugMessage( sMessage );

              }

              pCmdCtxt->nSndSeqState = DCC_SEQ_SENT;
              bDone = TRUE;       /* OK: successful command */  
            }           /* end if (DCPT_ScanForResponse) */

            /* extra handling invoked in all cases (error or not */
            DCPT_PerformExtraHandling(hWnd, pCmdCtxt, pCmdCtxt->bDisplay );
  
          }         /* if ( !DCPT_SendCommand( */

        }         /* end while !bDone */

      }         /* end normal processing */

      /* update xDccCtxt.nNbPendingSndMsg */
 
      EnterCriticalSection( &xDccCtxt.xCriticalSection );
      xDccCtxt.xSndCmd.nNbPendingCmd--;

      /* DGI_COMP_LOCAL_OK is a special code for responses processed locally */
      /* with no need to be forwarded to CDM-Rail.                           */
      /* Commands in state DGI_COMP_LOCAL_OK are to be removed from command  */
      /* list.                                                               */

      if ( pCmdCtxt->nCompletionCode != DGI_COMP_LOCAL_OK )
        pCmdCtxt->nSndSeqState = DCC_SEQ_DONE;
      else
      { /* detach will be done in GetNextExpected */ 
        pCmdCtxt->nSndSeqState = DCC_SEQ_LOCAL_DONE;
      }

      LeaveCriticalSection( &xDccCtxt.xCriticalSection );
    }           /* if ( xDccCtxt.nNbPendingSndCmd > 0 ) */

    else
    {
      if ( DCPT_LookForUnexpectedMessage( hWnd, xDccCtxt.hComPort,
                &xRcvCommandCtxt, FALSE  ) )
      {
      }         /* end if ( DCPT_LookForUnexpectedResponse */
    }

  }             /* end while ( 1 ) */

  return lReturn;
}

/*======================================================================*/
/*      DCPT_SendCommand                                                */

BOOL
DCPT_SendCommand( HWND                  hWnd,
                HANDLE                  hCom,
                DCC_COMMAND_CTXT        *pSndCommandCtxt,
                BOOL                    bShowCommand )
{
  unsigned char                 xSndByte[MAX_NB_CMD_BYTES];
  int                           i;
  DWORD                         nNbBytes, nNbBytesWritten;
  int                           nNbArgs;
  unsigned char                 sSubstring[10];                  
  char                          sMessage[128];
  BOOL                          bOK = TRUE;

  /* reset array */
  for ( i = 0; i < MAX_NB_CMD_BYTES; i++ )
    xSndByte[i] = '\0';

  nNbArgs = pSndCommandCtxt->xSndCommand.nNbArgs;

  // opening bracket, and command symbol
  sprintf( xSndByte , "<%c", pSndCommandCtxt->xSndCommand.cCommandCode );

  // build argument string
  for ( i = 0; i < nNbArgs ; i++ )
  {
    sprintf( sSubstring, " %d", pSndCommandCtxt->xSndCommand.xDataByte[i] );
    strcat( xSndByte, sSubstring );
  }

  // and close bracket
  strcat( xSndByte, ">" );

  nNbBytes = strlen( xSndByte );
    
  /* write buffer */
  if ( bOK )
  {
    if ( bShowCommand )
    {
      sprintf( sMessage, "Write %d bytes: %s", nNbBytes, xSndByte );
      MessageBox( hWnd, sMessage, "",
          MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
    }

    if ( !DCPU_WriteFile( hCom, xSndByte, nNbBytes, &nNbBytesWritten, TRUE ) ||
         nNbBytesWritten != nNbBytes )
      bOK = FALSE;
  }   

  if ( pDebugFile != NULL )
  {
    sprintf( sMessage , "send cmd %d %08X (%ld): %s  --> OK: %d", 
                        pSndCommandCtxt->nSndCommand, 
                        (unsigned long) pSndCommandCtxt,
                        pSndCommandCtxt->lIndex,
                        xSndByte, bOK );
    DCPU_PrintDebugMessage( sMessage );
  }

  return bOK;
}

/*======================================================================*/
/*      DCPT_ReadCommand                                                */
/*      Function similar to DCPI_ReadCommand, except that arguments     */
/*      RCV_CMD_MNEMONIC pCommand and pDataArray are replaced by a      */
/*      pointer to a whole DCC_COMMAND structure.                       */
/*      This function receives a DCC command.                           */
/*      - If bExpected is TRUE, the tempo is defined by nLatency.       */
/*      - If bExpected is FALSE, then the tempo is reset to 0.          */
/*      RETURN: error code.                                             */

RD_ERROR
DCPT_ReadCommand( HWND                  hWnd,
                HANDLE                  hCom,
                DCC_COMMAND             *pRcvCommand,
                unsigned char           xRcvByte[],
                int                     nLatency,
                RCV_CMD_MNEMONIC        nExpectedResponse,      // on CDM side
                unsigned char           *sExpectedRespHeader,   // from DCC++
                BOOL                    bExpected,
                BOOL                    bShowCommand,
                BOOL                    bStoreToFifo )
{
  int                           nByte;
  int                           nData;
  DWORD                         nNbBytesRead;
  int                           nTotalNbBytesRead = 0;
  char                          sMessage[2 * MAX_MSG_LEN];
  RD_ERROR                      nErrorCode = RD_ERR_NONE;
  BOOL                          bTimeOutModified = FALSE;

  /* reset array */
  for ( nByte = 0; nByte < MAX_NB_CMD_BYTES; nByte++ )
    xRcvByte[nByte] = '\0';
  for ( nData = 0; nData < MAX_NB_CMD_DATA; nData++ )
    pRcvCommand->xDataByte[nData] = 0;

  pRcvCommand->nResponse[0] = RCV_NONE;

  /* READ NEXT COMMAND FROM IF */
  /*===========================*/

  if ( nLatency != 0 && nLatency != ( DATA_COM_TIME_OUT / T_100_MS ) )
  {
    DCPU_SetTimeOut( hCom, nLatency * T_100_MS );
    bTimeOutModified = TRUE;
  }

  if ( !DCPU_ReadFile( hCom, xRcvByte, &nNbBytesRead, 
                       bExpected? RD_RESPONSE : RD_FEEDBACK, bStoreToFifo ) )
  {
    if ( nNbBytesRead == 0 )
      nErrorCode = RD_ERR_NO_RESPONSE;
    else
      nErrorCode = RD_ERR_ARG_FORMAT_ERROR;
  }
  else      /* read data successful */
  {        
    nTotalNbBytesRead = nNbBytesRead;
    nErrorCode = DCPT_LookForErrorCode( xRcvByte, nNbBytesRead );

    if ( nErrorCode == RD_ERR_NONE )
    {
      nErrorCode = DCPT_LookForUnexpectedResponse( pRcvCommand, 
                                                   sExpectedRespHeader,
                                                   xRcvByte, 
                                                   nTotalNbBytesRead );
    }

    if ( nErrorCode == RD_ERR_UNEXPECTED_RESPONSE )     // feedback
    {
      if ( pDebugFile != NULL )
      {
        sprintf( sMessage , "FEEDBACK: %s Expected header %s", 
                                        xRcvByte, sExpectedRespHeader );
        DCPU_PrintDebugMessage( sMessage );
      }
    }
    
    if ( nErrorCode == RD_ERR_NONE )
    {
      nErrorCode = DCPT_GetResponse( pRcvCommand, 
                                     nExpectedResponse,
                                     xRcvByte, 
                                     nTotalNbBytesRead );
    }
  }

  if ( bTimeOutModified )
    DCPU_SetTimeOut( hCom, DATA_COM_TIME_OUT );

  if ( pDebugFile != NULL && nNbBytesRead > 0 )
  {
     sprintf( sMessage, "%d read bytes: %s" , nNbBytesRead, xRcvByte ); 
     DCPU_PrintDebugMessage( sMessage );
  }

  if ( bShowCommand )
  {
    MessageBox(hWnd, xRcvByte, "", 
                      MB_APPLMODAL | MB_OK | MB_ICONINFORMATION);
  }   

  return nErrorCode;
}

/*======================================================================*/
/*      DCPT_LookForErrorCode                                           */
/*      This function looks for an error code in the received response. */
/*      In DCC++ the error code generated is <X>.                       */   

RD_ERROR
DCPT_LookForErrorCode( unsigned char *xRcvByte, int nNbBytes )
{
  RD_ERROR      nError;
  BOOL          bFound;
  int           i;

  if ( xRcvByte[0] != '<' )
    nError = RD_ERR_ARG_FORMAT_ERROR;
  else  // look for first non-blank char 
  {
    i = 1;
    bFound = FALSE;
    while ( i < nNbBytes && !bFound )
    {
      if ( xRcvByte[i] == ' ' )
        i++;
      else
        bFound = TRUE;
    }

    if ( bFound )
    {
      if ( xRcvByte[i] == 'X' || xRcvByte[i] == 'x' )
        nError = RD_ERR_UNEXPECTED_RESPONSE;
      else
        nError = RD_ERR_NONE;
    }
  }
  return nError;

}

/*======================================================================*/
/*      DCPT_LookForUnexpectedResponse                                  */
/*      This function looks if the received message id one of the       */
/*      DCC++ feedbacks.                                                */
/*      <p2> <p3> POWER OFF on short circuit.                           */
/*      <Q...> or <q....> sensor feedback.                              */
/*      <Y...> S88 feedback.                                            */

RD_ERROR
DCPT_LookForUnexpectedResponse( DCC_COMMAND     *pRcvCommand,
                                unsigned char   *sExpectedRespHeader,
                                unsigned char   *xRcvByte, 
                                int             nNbBytes )
{
  RD_ERROR      nErrorCode = RD_ERR_NONE;
  BOOL          bFound;
  int           i;
  char          sMessage[128];

  if ( xRcvByte[0] != '<' )     
    nErrorCode = RD_ERR_ARG_FORMAT_ERROR;
  else if ( strncmp( &xRcvByte[1] , "p0", 2 ) == 0 )       // power off
  {
    // may be response to command <0>
    if ( strncmp( sExpectedRespHeader, "p0", 2 ) == 0 )
      nErrorCode = RD_ERR_NONE;       // expected response
    else
    {
      pRcvCommand->nResponse[0] = RCV_NORMAL_OPER_RESUMED;
      nErrorCode = RD_ERR_UNEXPECTED_RESPONSE;
    }
  }

  else if ( strncmp( &xRcvByte[1] , "p1", 2 ) == 0 )  // resume operations
  {
    // may be response to command <1>
    if ( strncmp( sExpectedRespHeader, "p1", 2 ) == 0 )
      nErrorCode = RD_ERR_NONE;       // expected response
    else
    {
      pRcvCommand->nResponse[0] = RCV_TRACK_POWER_OFF;
      nErrorCode = RD_ERR_UNEXPECTED_RESPONSE;
    }
  }

  else if ( strncmp( &xRcvByte[1] , "p2", 2 ) == 0 )  // resume operations
  {
    pRcvCommand->nResponse[0] = RCV_SHORT_CIRCUIT_MAIN;
    nErrorCode = RD_ERR_UNEXPECTED_RESPONSE;
  }

  else if ( strncmp( &xRcvByte[1] , "p3", 2 ) == 0 )  // resume operations
  {
    pRcvCommand->nResponse[0] = RCV_SHORT_CIRCUIT_PROG;
    nErrorCode = RD_ERR_UNEXPECTED_RESPONSE;
  }

  else if ( xRcvByte[1] == 'Q' )
  {
    pRcvCommand->nResponse[0] = RCV_SENSOR_TRIGGERED;
    nErrorCode = RD_ERR_UNEXPECTED_RESPONSE;
  }

  else if ( xRcvByte[1] == 'q' )
  {
    pRcvCommand->nResponse[0] = RCV_SENSOR_RELEASED;
    nErrorCode = RD_ERR_UNEXPECTED_RESPONSE;
  }

  else if ( xRcvByte[1] == 'y' )
  {
    if ( sExpectedRespHeader[0] == 'y' )
      nErrorCode = RD_ERR_NONE;       // expected response
    else
    {
      pRcvCommand->nResponse[0] = RCV_S88_STATE;
      nErrorCode = RD_ERR_UNEXPECTED_RESPONSE;
    }
  }

  return nErrorCode;

}

/*======================================================================*/
/*      DCPT_GetResponse                                                */
/*      This function:                                                  */
/*      - searches for ending prompt '>',                               */
/*      - if not found, returns error RD_NO_PROMPT.                     */
/*      - sets index to 2 chars before this prompt.                     */
/*      - If there is no data:                                          */
/*        * if expected response if RCV_ACK: OK.                        */
/*        * if expected response if NOT RCV_ACK, return error:          */
/*                                              RD_ERR_NO_RESPONSE  */
/*      - If response is string "OK" (only in the of a write CV)        */
/*        * if expected response if RCV_ACK: OK.                        */
/*        * if expected response if NOT RCV_ACK, return error:          */
/*                                      RD_ERR_UNEXPECTED_RESPONSE  */
/*      - Else : a numeric value is expected. Since this numeric value  */
/*        may be prefixed par '=', this equal signal should be discarded*/
/*        * discard '=' if existing.                                    */


RD_ERROR
DCPT_GetResponse( DCC_COMMAND           *pSndCommand,
                RCV_CMD_MNEMONIC        nExpectedResponse,
                unsigned char           *xRcvBuffer, 
                int                     nNbBytes )
{
  RD_ERROR              nErrorCode = RD_ERR_NONE;
  int                   nDummyNb1, nDummyNb2, nCVNum, nCVVal;
  int                   i;
  char                  sMessage[128];

  xRcvBuffer[nNbBytes] = '\0';
  
  switch( nExpectedResponse )
  {
  case RCV_S88_STATE: 
    // merely take new state into account.
    DCPS_ExtractS88State( xRcvBuffer );        
    break;

  case RCV_SRVC_MODE_RESP_DIRECT_CV:    // pseudo response on SND_DIRECT_CV_READ
    // all non numeric chzracters by spaces
    for ( i = 0; i < strlen( xRcvBuffer ); i++ )
    {
      if ( xRcvBuffer[i] < (unsigned char) 0x30 || 
           xRcvBuffer[i] > (unsigned char) 0x39 )
        xRcvBuffer[i] = ' ';
    }
 
    // here the remaining string id like "1 1 3 64" (3 = CV num 64 = CV Val
    if ( sscanf( xRcvBuffer, "%d %d %d %d" , &nDummyNb1, &nDummyNb2, 
                                             &nCVNum, &nCVVal ) != 4 )
      nErrorCode = RD_ERR_CV_DIRECT_ERROR;
    else
    {
      if ( pDebugFile != NULL )
      {
        sprintf( sMessage , "DCPT_GetResponse nCVNum = %d nCVVal = %d" , nCVNum, nCVVal );
        DCPU_PrintDebugMessage( sMessage );
      }

      if ( nCVVal < 0 || nCVVal > 255 )
      {
        nErrorCode = RD_ERR_CV_DIRECT_ERROR;
      }
      else
        cCurrentCVValue = (unsigned char ) nCVVal;
    }
      
    break;
  }

  return nErrorCode;

}

/*======================================================================*/
/*      DCPT_ScanForResponse                                            */
/*      This function repeatedly reads the COM port until it reads the  */
/*      expected response to previously sent command.                   */

BOOL
DCPT_ScanForResponse(   HWND                    hWnd, 
                        HANDLE                  hCom, 
                        DCC_COMMAND_CTXT        *pSndCommandCtxt, 
                        BOOL                    bShow)
{
  DCC_COMPLETION_CODE   nCompletionCode = DGI_COMP_NONE;
  RD_ERROR              nRdErrorCode;
  RCV_CMD_MNEMONIC      nRdCommand = RCV_NONE;
  unsigned char         xRcvBuffer[MAX_NB_CMD_BYTES];
  char                  sMessage[64];

  while ( nCompletionCode == DGI_COMP_NONE )    // no response expected from DCC++
  {
    if ( pSndCommandCtxt->xSndCommand.nResponse[0] == RCV_NONE )
    {
      nRdErrorCode = RD_ERR_NONE;
      nRdCommand = RCV_ACK;
      pSndCommandCtxt->nRdErrorCode       = nRdErrorCode;
      pSndCommandCtxt->nRdCommand         = nRdCommand;
    }
    
    else
    {
      
      nRdErrorCode = 
        DCPT_ReadCommand( hWnd, hCom, &pSndCommandCtxt->xRcvCommand, 
                        xRcvBuffer,
                        pSndCommandCtxt->xSndCommand.nLatency, 
                        pSndCommandCtxt->xSndCommand.nResponse[0],
                        pSndCommandCtxt->xSndCommand.sExpectedRespHeader,
                        TRUE, bShow,
                        TRUE );
      nRdCommand = pSndCommandCtxt->xSndCommand.nResponse[0];
      pSndCommandCtxt->nRdErrorCode       = nRdErrorCode;
      pSndCommandCtxt->nRdCommand         = nRdCommand;
    }

    if ( nRdErrorCode == RD_ERR_UNEXPECTED_RESPONSE )
    {
      // process unexpected message, and leave nCompletion to DGI_COMP_NONE
      // as to reiterate for next read.
      DCPT_HandleUnexpectedMessage( pSndCommandCtxt, 
                                    pSndCommandCtxt->xRcvCommand.nResponse[0],
                                    xRcvBuffer,
                                    FALSE );
    }

    else if ( nRdErrorCode != RD_ERR_NONE )
    {
      DCPI_DisplayReadError( hWnd, nRdErrorCode, "" );
      nCompletionCode = pSndCommandCtxt->nCompletionCode = DGI_COMP_READ_ERR;
      xDccCtxt.xServerDesc.nLastRdErrorCode = nRdErrorCode;
      if ( nRdErrorCode == RD_ERR_UNRECOGNIZED_COMMAND )
        DCPT_DisplayErrorMessage( hWnd, pSndCommandCtxt );      
    }
    else                /* no error */
    {
      switch( nRdCommand )
      {
      /*
      case RCV_SRVC_MODE_RESP_DIRECT_CV:
        nCompletionCode = pSndCommandCtxt->nCompletionCode = DGI_COMP_OK;
        DCPU_UpdateLastCmdFifoEntry( pSndCommandCtxt );
        break;
      */

      default:    /* standard commands */
        nCompletionCode = pSndCommandCtxt->nCompletionCode = DGI_COMP_OK;
        DCPU_UpdateLastCmdFifoEntry( pSndCommandCtxt );
        break;
      }         /* end switch */

    }

  }   /* end while ( nCompletionCode ... ) */

  return ( nCompletionCode == DGI_COMP_OK );
}


/*======================================================================*/
/*      DCPT_PerformStandaloneHandling                                  */
/*      Performs handling of commands which are not handled by the SPROG*/

BOOL
DCPT_PerformStandaloneHandling( HWND            hWnd, 
                           DCC_COMMAND_CTXT     *pSndCommandCtxt, 
                           BOOL                 bShow)
{
  DCC_COMMAND           *pRcvCommand;
  unsigned char         cBuffer[10];
  char                  sMessage[128];

  if ( pDebugFile != NULL )
  {
    sprintf( sMessage , "standalone handling cmd %d (%ld)", 
                        pSndCommandCtxt->nSndCommand, pSndCommandCtxt->lIndex );
    DCPU_PrintDebugMessage( sMessage );
  }

  pRcvCommand = &pSndCommandCtxt->xRcvCommand;

  switch( pSndCommandCtxt->nSndCommand )
  {
  case SND_STOP_ALL_LOCOS_REQUEST:
  case SND_ENTER_SERVICE_MODE:
  case SND_EXIT_SERVICE_MODE:
  case SND_ACCESSORY_OPER:      // release only!
    pSndCommandCtxt->nRdErrorCode       =   RD_ERR_NONE;
    pSndCommandCtxt->nRdCommand         =   RCV_ACK;
    pSndCommandCtxt->nCompletionCode    =   DGI_COMP_OK; 
    break;

  case SND_ACCESSORY_INFO_REQUEST:
    pSndCommandCtxt->nRdErrorCode       =   RD_ERR_NONE;
    pSndCommandCtxt->nRdCommand         =   RCV_ACCESSORY_INFO;
    pSndCommandCtxt->nCompletionCode    =   DGI_COMP_OK;
    // data retrieved by MapResponseToGeneric,  on DCPT_GetNextExpectedResponse
    break;

  case SND_DETECTOR_INFO_REQUEST:
    pSndCommandCtxt->nRdErrorCode       =   RD_ERR_NONE;
    pSndCommandCtxt->nRdCommand         =   RCV_DETECTOR_INFO;
    pSndCommandCtxt->nCompletionCode    =   DGI_COMP_OK;
    // data retrieved by MapResponseToGeneric,  on DCPT_GetNextExpectedResponse
    break;

  case SND_LOCO_INFO_REQUEST:
    pSndCommandCtxt->nRdErrorCode       =   RD_ERR_NONE; 
    pSndCommandCtxt->nRdCommand         =   RCV_NORMAL_LOCO_INFO;
    pSndCommandCtxt->nCompletionCode    =   DGI_COMP_OK; 
    // data retrieved by MapResponseToGeneric,  on DCPT_GetNextExpectedResponse
    break;

  case SND_CMDSTATION_STATUS_REQUEST:
    pSndCommandCtxt->nRdErrorCode       = RD_ERR_NONE;
    pSndCommandCtxt->nRdCommand         = RCV_CMDSTATION_STATUS;
    pSndCommandCtxt->nCompletionCode    = DGI_COMP_OK; 
    if ( pDebugFile != NULL )
    {
      sprintf( sMessage , "Command station status request cmd %d (%ld) response %d", 
        pSndCommandCtxt->nSndCommand, 
        pSndCommandCtxt->lIndex,
        pSndCommandCtxt->nRdCommand );
      DCPU_PrintDebugMessage( sMessage );
    }
    break;

  case SND_SERVICE_MODE_EVENT_REQUEST:
    pSndCommandCtxt->nRdErrorCode       = nCurrentCVError;
    pSndCommandCtxt->nRdCommand         = RCV_SRVC_MODE_RESP_DIRECT_CV;
    pSndCommandCtxt->nCompletionCode    = DGI_COMP_OK; 
    pRcvCommand->xDataByte[0]           = cCurrentCVValue;
  }


  cBuffer[0] = pSndCommandCtxt->nSndCommand;
  DCPU_CopyBufferToFifo( cBuffer, 1, (char) 0xDD, 
                         TRUE,  // bSend
                         TRUE,  // bNewCmd
                         CMD_STANDALONE );

}

/*======================================================================*/
/*      DCPT_PerformExtraHandling                                       */
/*      Performs additonal handling in some special cases like:         */
/*      emergency stop                                                  */
/*      power off                                                       */
/*      resume operations                                               */

void
DCPT_PerformExtraHandling( HWND                 hWnd, 
                           DCC_COMMAND_CTXT     *pSndCommandCtxt, 
                           BOOL                 bShow)
{
  char                  sMessage[128];

  switch( pSndCommandCtxt->nSndCommand )
  {
  case SND_RESUME_OPER_REQUEST:
    // if ( xDccCtxt.xServerDesc.bEmergencyStop )
    //  DCPS_RestartAllLocos();
    DCPT_GenerateEventMessage( RCV_NORMAL_OPER_RESUMED );
    xDccCtxt.xServerDesc.bEmergencyStop = FALSE;
    xDccCtxt.xServerDesc.bPowerOff = FALSE;
    break;

  case SND_STOP_ALL_OPERS_REQUEST:
    DCPT_GenerateEventMessage( RCV_TRACK_POWER_OFF );
    xDccCtxt.xServerDesc.bPowerOff = TRUE;
    break;

  case SND_STOP_ALL_LOCOS_REQUEST:
    DCPT_GenerateEventMessage( RCV_EMERGENCY_STOP );
    xDccCtxt.xServerDesc.bEmergencyStop = TRUE;
    break;

  case SND_DETECTOR_INFO_REQUEST:
    // this command is handled as standalone with respect to CDM-Rail 
    // But (in parallel) a refresh command Y is sent to DCC++, as to update
    // the S88, for next state request.
    DCPS_S88UpdateRequest();
    break;

  case SND_DIRECT_CV_READ:    
    // whatever the result: make return OK 
    // real result will returned by SND_SERVICE_MODE_EVENT_REQUEST
    pSndCommandCtxt->nRdCommand         = RCV_ACK;
    pSndCommandCtxt->nCompletionCode    = DGI_COMP_OK;
    nCurrentCVError = pSndCommandCtxt->nRdErrorCode;
    if ( pSndCommandCtxt->nRdErrorCode == RD_ERR_NONE )
    {
      // cCurrentCVValue has been extracted in DCPT_GetResponse 
        
    }
    else
    {
      nCurrentCVError = RD_ERR_UNEXPECTED_RESPONSE;
      cCurrentCVValue = (unsigned char) 0;
      // make return without error. Error will be on service result
      pSndCommandCtxt->nRdErrorCode = RD_ERR_NONE;
    }

    pSndCommandCtxt->nRdCommand         = RCV_ACK;

    if ( pDebugFile != NULL )
    {
      sprintf( sMessage , "extra handling cmd %d (%ld) error %d cvalue %02X", 
                        pSndCommandCtxt->nSndCommand, pSndCommandCtxt->lIndex,
                        nCurrentCVError,
                        pSndCommandCtxt->xRcvCommand.xDataByte[0]  );
      DCPU_PrintDebugMessage( sMessage );
    }
    break;

  }
}

/*======================================================================*/
/*      DCPT_GenerateEventMessage                                       */
/*      This function creates a new DCC_COMMAND_CTXT, appended to list  */
/*      unexpected messages, with read command ndCommand.               */
/*      Used for broadcast commands like STOP, POFF, RESUME.            */
/*                                                                      */
/*      RETURN: TRUE if broadcast command, FALSE if not.                */

BOOL    
DCPT_GenerateEventMessage( RCV_CMD_MNEMONIC nRdCommand )
{
  BOOL                  bOK = TRUE;
  DCC_COMMAND_CTXT      *pRcvMsgCtxt;
  unsigned char         cBuffer[10];

  if ( nRdCommand != RCV_ACK )
  {
    EnterCriticalSection( &xDccCtxt.xCriticalSection );
    if ( xDccCtxt.xRcvMsg.nNbCmd >= MAX_RCV_MSG )
    {
      xDccCtxt.lNbRcvOverflows++;
      DCPU_DetachCommandCtxt( &xDccCtxt.xRcvMsg );
    }
    pRcvMsgCtxt = DCPU_NewCommandCtxt( &xDccCtxt.xRcvMsg, PRI_0 );
    pRcvMsgCtxt->nRdCommand = cBuffer[0] = nRdCommand;
    pRcvMsgCtxt->nCompletionCode = DGI_COMP_OK;
    pRcvMsgCtxt->nCommandType = CMD_LOCAL_FEEDBACK;
    DCPU_CopyBufferToFifo( cBuffer, 1, (char) 0xDD, 
                           FALSE,
                           TRUE,
                           CMD_LOCAL_FEEDBACK );

    xDccCtxt.lTotalNbMsgReceived++;

    LeaveCriticalSection( &xDccCtxt.xCriticalSection );
  }

  return bOK;
}

/*======================================================================*/
/*      DCCT_HandleUnexpectedMessage                                    */
/*      This function creates a new DCC_COMMAND_CTXT, appended to list  */
/*      unexpected messages, and stores received command in that new    */
/*      DCC_COMMAND_CTXT.                                               */
/*      RCV_ACK messages ARE IGNORED.                                   */
/*                                                                      */
/*      RETURN: TRUE if broadcast command, FALSE if not.                */

BOOL    
DCPT_HandleUnexpectedMessage(   DCC_COMMAND_CTXT        *pCmdCtxt,
                                RCV_CMD_MNEMONIC        nRdCommand,
                                unsigned char           xRcvByte[],
                                BOOL                    bOutCommandCycle )
{
  BOOL                  bOK = TRUE;
  DCC_COMMAND_CTXT      *pRcvMsgCtxt;
  char                  *pCommandStr = "\0";
  char                  sMessage[128];

  switch( nRdCommand )
  {
  case RCV_NORMAL_OPER_RESUMED:
    DCPT_GenerateEventMessage( RCV_NORMAL_OPER_RESUMED );
    break;

  case RCV_TRACK_POWER_OFF:
  case RCV_SHORT_CIRCUIT_MAIN:
  case RCV_SHORT_CIRCUIT_PROG:
    DCPT_GenerateEventMessage( RCV_TRACK_POWER_OFF );
    break;

  case RCV_SENSOR_TRIGGERED:
  case RCV_SENSOR_RELEASED:     // currently unhandled
    break;

  case RCV_S88_STATE:
    if ( pDebugFile != NULL )
    {
      sprintf( sMessage , "Hangle U M" );
      DCPU_PrintDebugMessage( sMessage );
    }


    if ( DCPS_ExtractS88State( xRcvByte ) &&
         DCPS_S88StateChanged() )     // extraction OK
    {
      DCPT_GenerateEventMessage( RCV_FEEDBACK_BROADCAST );
    if ( pDebugFile != NULL )
    {
      sprintf( sMessage , "feedback forwarded" );
      DCPU_PrintDebugMessage( sMessage );
    }
    }
    break;

  }


  return bOK;
}





/*======================================================================*/
/*      DCPT_LookForUnexpectedMessage                                   */

BOOL
DCPT_LookForUnexpectedMessage(   
                        HWND                    hWnd, 
                        HANDLE                  hCom, 
                        DCC_COMMAND_CTXT        *pRcvCommandCtxt, 
                        BOOL                    bShow)
{
  DWORD                 nNbExpBytes = 10;
  DWORD                 nNbBytesRead;
  unsigned char         xRcvBytes[MAX_NB_CMD_BYTES];
  RD_ERROR              nRdErrorCode;
  BOOL                  bOK = FALSE;

  // set (almost) immediate read
  //DCPU_SetTimeOut( hCom, (DWORD) 0 );  // Done in DCPU_ReadFile()

  
  /*
  if ( DCPU_ReadFile( hCom, xRcvBytes, &nNbBytesRead, RD_FEEDBACK, TRUE) )
  {
    bOK = TRUE;
  }
  */

  nRdErrorCode = 
      DCPT_ReadCommand( hWnd, hCom, 
                        &pRcvCommandCtxt->xRcvCommand,
                        xRcvBytes,
                        0,      // to avoid latency modif at DCPT_ReadCommand level
                        RCV_NONE,
                        "___",          // dummy string to avoid any match
                        FALSE,          // not expected 
                        bShow,   
                        TRUE );         // store in FIFO

  if ( nRdErrorCode == RD_ERR_UNEXPECTED_RESPONSE )     // normal situation
  {
    bOK = TRUE;
    // process feedback message, 
    DCPT_HandleUnexpectedMessage( pRcvCommandCtxt, 
                                  pRcvCommandCtxt->xRcvCommand.nResponse[0],
                                  xRcvBytes,
                                  TRUE );       // OutCommandcycle

  }
  else
  {
  }    // ignored

  // restore time-out 
  DCPU_SetTimeOut( hCom, DATA_COM_TIME_OUT );

  return bOK;
}


/*======================================================================*/
/*      DCPT_DisplayErrorMessage                                        */
/*      This function is invoked whenever a comm error is detected.     */
/*      If global flag bDisplayErrors is set,                           */
/*      It displays information on the error, and the dumps the 64 last */
/*      bytes received + transmit.                                      */

void
DCPT_DisplayErrorMessage( HWND hWnd, DCC_COMMAND_CTXT *pCommandCtxt )
{
  char                  sMessage[MAX_LONG_MSG_LEN];

  if ( bDisplayErrors )
  {
    sprintf( sMessage, "ERROR!!\n" 
                       "Completion code : %s\n"
                       "Error code:       %s\n"
                       "Read command:     %s\n",
    DCPU_CompletionCodeStr( pCommandCtxt->nCompletionCode ),
    DCPU_RdErrorStr( pCommandCtxt->nRdErrorCode ),
    DCPU_RcvCmdMnemonicStr( pCommandCtxt->nRdCommand ) );
    MessageBox( hWnd, sMessage, "INTERFACE THREAD", 
                MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );

    DCPU_DumpDbgFifo( hWnd );
  }

  if ( pDumpFile != NULL )
  {
    fprintf( pDumpFile, "ERROR!!\n" 
                        "time : %ld\n"
                        "Completion code : %s\n"
                        "Error code:       %s\n"
                        "Read command:     %s\n",
      xDccCtxt.xServerDesc.lCurTimeTick,
      DCPU_CompletionCodeStr( pCommandCtxt->nCompletionCode ),
      DCPU_RdErrorStr( pCommandCtxt->nRdErrorCode ),
      DCPU_RcvCmdMnemonicStr( pCommandCtxt->nRdCommand ) );
    fflush( pDumpFile );
  }
}

